# Copyright © 2016–2026 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Simple recursive-descent parser.

This module implements the following class:

* Parser: Recursive-descent parser.
"""
import re
from typing import Callable, Optional, Pattern
from .exceptions import EndOfInput, UnexpectedInput
from .typealiases import YangIdentifier

# Local type aliases
TransitionTable = list[dict[str, Callable[[], int]]]
"""Transition table for a DFA."""


class Parser:

    """Recursive-descent parser with support for YANG syntactic elements."""

    # Regular expressions

    ident_re = re.compile("[a-zA-Z_][a-zA-Z0-9_.-]*")
    """Regular expression for YANG identifier."""

    ws_re = re.compile(r"[ \n\t\r]*")
    """Regular expression for whitespace."""

    _uint = "[0-9]+"
    uint_re = re.compile(_uint)
    """Regular expression for unsigned integer."""

    ufloat_re = re.compile(fr"{_uint}(\.{_uint})?|\.{_uint}")
    """Regular expression for unsigned float."""

    def __init__(self, text: str):
        """Initialize the class instance.

        Args:
            text: Input text.
        """
        self.input = text
        """Input text for the parser."""
        self.offset = 0
        """Current position in the input text."""

    def __str__(self) -> str:
        """Return string representation of the receiver's input text and state.
        """
        return self.input[:self.offset] + "§" + self.input[self.offset:]

    def adv_skip_ws(self) -> bool:
        """Advance offset and skip optional whitespace."""
        self.offset += 1
        return self.skip_ws()

    def at_end(self) -> bool:
        """Return ``True`` if at end of input."""
        return self.offset >= len(self.input)

    def at_last_char(self) -> bool:
        """Return ``True`` if at last character of input."""
        return self.offset == (len(self.input) - 1)

    def char(self, c: str) -> None:
        """Parse the specified character.

        Args:
            c: One-character string.

        Raises:
            EndOfInput: If past the end of `self.input`.
            UnexpectedInput: If the next character is different from `c`.
        """
        if self.peek() == c:
            self.offset += 1
        else:
            raise UnexpectedInput(self, f"char '{c}'")

    def dfa(self, ttab: TransitionTable, init: int = 0) -> int:
        """Run a DFA and return the final (negative) state.

        Args:
            ttab: Transition table (with possible side-effects).
            init: Initial state.

        Raises:
            EndOfInput: If past the end of `self.input`.
        """
        state = init
        while True:
            disp = ttab[state]
            ch = self.peek()
            state = disp.get(ch, disp[""])()
            if state < 0:
                return state
            self.offset += 1

    def line_column(self) -> tuple[int, int]:
        """Return line and column coordinates."""
        ln = self.input.count("\n", 0, self.offset)
        c = (self.offset if ln == 0 else
             self.offset - self.input.rfind("\n", 0, self.offset) - 1)
        return (ln + 1, c)

    def match_regex(self, regex: Pattern, required: bool = False,
                    meaning: str = "") -> str:
        """Parse input based on a regular expression .

        Args:
            regex: Compiled regular expression object.
            required: Should the exception be raised on unexpected input?
            meaning: Meaning of `regex` (for use in error messages).

        Raises:
            UnexpectedInput: If no syntactically correct keyword is found.
        """
        mo = regex.match(self.input, self.offset)
        if mo:
            self.offset = mo.end()
            return mo.group()
        if required:
            raise UnexpectedInput(self, meaning)

    def one_of(self, chset: str) -> str:
        """Parse one character form the specified set.

        Args:
            chset: string of characters to try as alternatives.

        Returns:
            The character that was actually matched.

        Raises:
            UnexpectedInput: If the next character is not in `chset`.
        """
        res = self.peek()
        if res in chset:
            self.offset += 1
            return res
        raise UnexpectedInput(self, "one of " + chset)

    def peek(self) -> str:
        """Return the next character without advancing offset.

        Raises:
            EndOfInput: If past the end of `self.input`.
        """
        try:
            return self.input[self.offset]
        except IndexError:
            raise EndOfInput(self)

    def prefixed_name(self) -> tuple[YangIdentifier, Optional[YangIdentifier]]:
        """Parse identifier with an optional colon-separated prefix."""
        i1 = self.yang_identifier()
        try:
            next = self.peek()
        except EndOfInput:
            return (i1, None)
        if next != ":":
            return (i1, None)
        self.offset += 1
        return (self.yang_identifier(), i1)

    def remaining(self) -> str:
        """Return the remaining part of the input string."""
        res = self.input[self.offset:]
        self.offset = len(self.input)
        return res

    def skip_ws(self) -> bool:
        """Skip optional whitespace."""
        return len(self.match_regex(self.ws_re)) > 0

    def test_string(self, string: str) -> bool:
        """If `string` comes next, return ``True`` and advance offset.

        Args:
            string: string to test
        """
        if self.input.startswith(string, self.offset):
            self.offset += len(string)
            return True
        return False

    def unsigned_integer(self) -> int:
        """Parse and return an unsigned integer."""
        return int(self.match_regex(self.uint_re, True, "unsigned integer"))

    def unsigned_float(self) -> float:
        """Parse and return unsigned floating-point number."""
        return float(self.match_regex(self.ufloat_re, True, "unsigned float"))

    def up_to(self, term: str) -> str:
        """Parse and return segment terminated by the first occurence
        of a string.

        Args:
            term: Terminating string.

        Raises:
            EndOfInput: If `term` does not occur in the rest of the input text.
        """
        end = self.input.find(term, self.offset)
        if end < 0:
            raise EndOfInput(self)
        res = self.input[self.offset:end]
        self.offset = end + 1
        return res

    def yang_identifier(self) -> YangIdentifier:
        """Parse and return YANG identifier.

        Raises:
            UnexpectedInput: If no syntactically correct keyword is found.
        """
        return self.match_regex(self.ident_re, True, "YANG identifier")



# Copyright © 2016–2026 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Extended XML parser that preserves the xmlns attributes
"""

import xml.etree.ElementTree as ET


class XMLParser(ET.XMLPullParser):
    '''
    Extended XML parser that add namespaces to ELements as
    xmlns/xmlns:... attributes
    '''
    def __init__(self, source: str = None):
        '''
        Initialize the XML parser

        Args:
            source: optional string containing the whole XML document
        '''
        super().__init__(events=['start', 'start-ns', 'end-ns', 'end'])

        self._root = None
        self._namespaces = list()

        if source:
            self.feed(source)
            self.close()
            self.parse()

    def parse(self):
        '''Parse all events in current available data'''
        for ev_type, ev_data in self.read_events():
            if ev_type == 'start-ns':
                ns_name, ns_url = ev_data
                self._namespaces.append((ns_name, ns_url))
            elif ev_type == 'end-ns':
                ns_name = self._namespaces.pop()

            elif ev_type == 'start' and self._root is None:
                self._root = ev_data
            elif ev_type == 'end':
                for ns_name, ns_url in self._namespaces:
                    attr = 'xmlns' if ns_name == '' else 'xmlns:'+ns_name
                    ev_data.attrib[attr] = ns_url

    @property
    def root(self):
        '''Return root node

        Only valid if the first event has been parsed'''
        return self._root



# Copyright © 2016–2026 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the entry point for a validation script."""

import argparse
import json
import os
import pickle
import sys
import importlib.metadata
from typing import Optional
from yangson import DataModel
from yangson.enumerations import ContentType, ValidationScope
from yangson.exceptions import (
    BadYangLibraryData, FeaturePrerequisiteError, InvalidArgument,
    MultipleImplementedRevisions, ModuleNotFound, ModuleNotRegistered,
    NonexistentSchemaNode, RawMemberError, RawTypeError,
    SchemaError, SemanticError, YangTypeError)
from yangson.typealiases import PrefName


def main(infile: Optional[str] = None, pickled: bool = False,
         path: Optional[str] = None,
         scope: ValidationScope = ValidationScope.all,
         ctype: ContentType = ContentType.all, set_id: bool = False,
         tree: bool = False, no_types: bool = False,
         digest: bool = False, dump: Optional[str] = None,
         subschema: Optional[PrefName] = None,
         validate: Optional[str] = None) -> int:
    """Entry-point for the command-line utility.

    Args:
        infile: Name of the input file with YANG library or pickled data model
        pickled: Interpret the input file as a pickled data model object
        path: Colon-separated list of directories to search  for YANG modules
        scope: Validation scope (syntax, semantics or all)
        ctype: Content type of the data instance (config, nonconfig or all)
        set_id: If `True`, print module set id
        tree: If `True`, print schema tree
        no_types: If `True`, don't print types in schema tree
        digest: If `True`, print schema digest
        dump: Dump pickled data model to the file given as argument
        subschema: Prefixed name of an RPC or notification subschema
        validate: Name of file to validate against the schema.

    Returns:
        Numeric return code (0=no error, 2=YANG error, 1=other)
    """
    if infile is None:
        parser = argparse.ArgumentParser(
            prog="yangson",
            description="Validate JSON data against a YANG data model.")
        parser.add_argument(
            "-V", "--version", action="version",
            help="show the Yangson library version number and exit",
            version=f"%(prog)s {importlib.metadata.version('yangson')}")
        parser.add_argument(
            "infile", metavar="INFILE",
            help=("file name with JSON-encoded YANG library [RFC 7895]"
                  " or pickled data model"))
        igrp = parser.add_mutually_exclusive_group()
        igrp.add_argument(
            "-p", "--path",
            help=("colon-separated list of directories to search"
                  " for YANG modules"))
        igrp.add_argument(
            "-P", "--pickled", action="store_true",
            help="interpret INFILE as pickled data model")
        ogrp = parser.add_mutually_exclusive_group()
        ogrp.add_argument(
            "-i", "--id", action="store_true",
            help="print module set id")
        ogrp.add_argument(
            "-t", "--tree", action="store_true",
            help="print schema tree as ASCII art")
        ogrp.add_argument(
            "-d", "--digest", action="store_true",
            help="print schema digest in JSON format")
        ogrp.add_argument(
            "-D", "--dump", metavar="FILE",
            help="dump pickled data model to FILE")
        ogrp.add_argument(
            "-v", "--validate", metavar="INST",
            help="name of file with JSON-encoded instance data")
        parser.add_argument(
            "-s", "--scope", choices=["syntax", "semantics", "all"],
            default="all", help="validation scope (default: %(default)s)")
        parser.add_argument(
            "-c", "--ctype", choices=["config", "nonconfig", "all"],
            default="all",
            help="content type of the data instance (default: %(default)s)")
        parser.add_argument(
            "-S", "--subschema",
            help="prefixed name of an RPC or notification subschema")
        parser.add_argument(
            "-n", "--no-types", action="store_true",
            help="suppress type info in tree output")
        args = parser.parse_args()
        infile = args.infile
        path = args.path
        pickled = args.pickled
        scope = ValidationScope[args.scope]
        ctype = ContentType[args.ctype]
        set_id = args.id
        tree = args.tree
        dump = args.dump
        no_types = args.no_types
        digest = args.digest
        subschema = args.subschema
        validate = args.validate
    if pickled:
        try:
            with open(infile, "rb") as pif:
                dm = pickle.load(pif)
        except (FileNotFoundError, pickle.UnpicklingError) as e:
            print("Pickle file:", str(e), file=sys.stderr)
            return 1
    else:
        try:
            with open(infile, encoding="utf-8") as ylf:
                yl = ylf.read()
        except (FileNotFoundError, PermissionError, UnicodeDecodeError,
                json.decoder.JSONDecodeError) as e:
            print("YANG library:", str(e), file=sys.stderr)
            return 1
        sp = path if path else os.environ.get("YANG_MODPATH", ".")
        try:
            dm = DataModel(yl, tuple(sp.split(":")))
        except BadYangLibraryData as e:
            print("Invalid YANG library:", str(e), file=sys.stderr)
            return 2
        except FeaturePrerequisiteError as e:
            print("Unsupported pre-requisite feature:", str(e),
                  file=sys.stderr)
            return 2
        except MultipleImplementedRevisions as e:
            print("Multiple implemented revisions:", str(e), file=sys.stderr)
            return 2
        except ModuleNotFound as e:
            print("Module not found:", str(e), file=sys.stderr)
            return 2
        except ModuleNotRegistered as e:
            print("Module not registered:", str(e), file=sys.stderr)
            return 2
    if set_id:
        print(dm.module_set_id())
        return 0
    if tree:
        print(dm.ascii_tree(no_types))
        return 0
    if digest:
        print(dm.schema_digest())
        return 0
    if dump:
        with open(dump, "wb") as pif:
            try:
                pickle.dump(dm, pif, pickle.HIGHEST_PROTOCOL)
                return 0
            except pickle.PicklingError as e:
                print("Pickling failed:", str(e), file=sys.stderr)
    if not validate:
        return 0
    try:
        with open(validate, encoding="utf-8") as instf:
            itxt = json.load(instf)
    except (FileNotFoundError, PermissionError, UnicodeDecodeError,
            json.decoder.JSONDecodeError) as e:
        print("Instance data:", str(e), file=sys.stderr)
        return 1
    try:
        i = dm.from_raw(itxt, subschema)
    except RawMemberError as e:
        print("Illegal object member:", str(e), file=sys.stderr)
        return 3
    except RawTypeError as e:
        print("Invalid type:", str(e), file=sys.stderr)
        return 3
    except InvalidArgument:
        print("Invalid subschema id:", subschema, "(missing prefix?)",
              file=sys.stderr)
        return 3
    except NonexistentSchemaNode:
        print("Subschema", subschema, "not found", file=sys.stderr)
        return 3
    try:
        i.validate(scope, ctype)
    except SchemaError as e:
        print("Schema error:", str(e), file=sys.stderr)
        return 3
    except SemanticError as e:
        print("Semantic error:", str(e), file=sys.stderr)
        return 3
    except YangTypeError as e:
        print("Invalid type:", str(e), file=sys.stderr)
        return 3
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright © 2016–2026 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""YANG schema nodes.

This module implements the following classes:

* SchemaNode: Abstract class for schema nodes.
* InternalNode: Abstract class for schema nodes that have children.
* GroupNode: Anonymous group of schema nodes.
* SchemaTreeNode: Root node of a schema tree.
* DataNode: Abstract class for data nodes.
* TerminalNode: Abstract class for schema nodes that have no children.
* ContainerNode: YANG container node.
* SequenceNode: Abstract class for schema nodes that represent a sequence.
* ListNode: YANG list node.
* ChoiceNode: YANG choice node.
* CaseNode: YANG case node.
* RpcActionNode: YANG rpc or action node.
* InputNode: YANG input node.
* OutputNode: YANG output node.
* NotificationNode: YANG notification node.
* LeafNode: YANG leaf node.
* LeafListNode: YANG leaf-list node.
* AnyContentNode: Abstract superclass for anydata and anyxml nodes..
* AnydataNode: YANG anydata node.
* AnyxmlNode: YANG anyxml node.
"""
from collections import deque
from collections.abc import MutableSet
from datetime import datetime
from itertools import product
from typing import Any, Optional
import xml.etree.ElementTree as ET
from .constraint import Must
from .datatype import DataType, LinkType, IdentityrefType
from .enumerations import (Axis, ContentType, DefaultDeny,
                           NodeStatus, ValidationScope)
from .exceptions import (
    AnnotationTypeError, InvalidArgument,
    MissingAnnotationTarget, MissingModuleNamespace, RawMemberError,
    RawTypeError, SchemaError, SemanticError, UndefinedAnnotation,
    YangsonException, YangTypeError)
from .instance import (ArrayEntry, InstanceNode, MemberName, ObjectMember)
from .instroute import InstanceRoute
from .instvalue import (
    ArrayValue, EntryValue, MetadataObject, ObjectValue, Value)
from .schemadata import IdentityAdjacency, SchemaContext, SchemaData
from .schpattern import (ChoicePattern, ConditionalPattern, Empty, Member,
                         NotAllowed, Pair, SchemaPattern)
from .statement import Statement
from .typealiases import (DataPath, InstanceName, JSONPointer, QualName,
                          RawEntry, RawList, RawObject, RawScalar, RawValue,
                          RawMetadataObject, ScalarValue, SchemaRoute,
                          YangIdentifier)
from .xpathast import Expr, LocationPath, Step, Root
from .xpathparser import XPathParser


class Annotation:
    """Class for metadata annotations [RFC 7952]."""

    def __init__(self, type: DataType, description: str = None):
        """Initialize the class instance."""
        self.type = type
        self.description = description


class SchemaNode:
    """Abstract class for all schema nodes."""

    def __init__(self):
        """Initialize the class instance."""
        self.name: Optional[YangIdentifier] = None
        """Name of the receiver."""
        self.ns: Optional[YangIdentifier] = None
        """Namespace of the receiver."""
        self.parent: Optional["InternalNode"] = None
        """Parent schema node."""
        self.description: Optional[str] = None
        """Description of the receiver."""
        self.must: list[Must] = []
        """List of "must" expressions attached to the receiver."""
        self.when: Optional["Expr"] = None
        """Optional "when" expression that makes the receiver conditional."""
        self.val_count = 0
        self._ctype = None
        """Content type of the receiver."""
        self._status = None
        """Status of node definition."""

    @property
    def qual_name(self) -> QualName:
        """Qualified name of the receiver."""
        return (self.name, self.ns)

    @property
    def config(self) -> bool:
        """Does the receiver (also) represent configuration?"""
        return self.content_type().value & ContentType.config.value != 0

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node in the complete data tree?"""
        return False

    @property
    def mandatory_config(self) -> bool:
        """Is the receiver mandatory node in a configuration?"""
        return self.mandatory and self.config

    @property
    def status(self) -> NodeStatus:
        """Return receiver's definition status."""
        return self._status if self._status else self.parent.status

    def delete(self) -> None:
        """Remove the receiver from the schema."""
        self.parent.children.remove(self)

    def schema_root(self) -> "SchemaTreeNode":
        """Return the root node of the receiver's schema."""
        sn = self
        while sn.parent:
            sn = sn.parent
        return sn

    def content_type(self) -> ContentType:
        """Return receiver's content type."""
        return self._ctype if self._ctype else self.parent.content_type()

    def data_parent(self) -> Optional["InternalNode"]:
        """Return the closest ancestor data node."""
        parent = self.parent
        while parent:
            if isinstance(parent, DataNode):
                return parent
            parent = parent.parent

    def iname(self) -> InstanceName:
        """Return the instance name corresponding to the receiver."""
        dp = self.data_parent()
        return (self.name if dp and self.ns == dp.ns
                else self.ns + ":" + self.name)

    def data_path(self) -> DataPath:
        """Return the receiver's data path."""
        dp = self.data_parent()
        return (dp.data_path() if dp else "") + "/" + self.iname()

    def state_roots(self) -> list[DataPath]:
        """Return a list of data paths to descendant state data roots."""
        return [r.data_path() for r in self._state_roots()]

    def from_raw(self, rval: RawValue, jptr: JSONPointer = "") -> Value:
        """Return instance value transformed from a raw value using receiver.

        Args:
            rval: Raw value.
            jptr: JSON pointer of the current instance node.

        Raises:
            RawMemberError: If a member inside `rval` is not defined in the
                schema.
            RawTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        raise NotImplementedError

    def from_xml(self, rval: ET.Element, jptr: JSONPointer = "",
                 isroot: bool = False) -> Value:
        """Return instance value transformed from a raw value using receiver.

        Args:
            rval: XML node.
            jptr: JSON pointer of the current instance node.

        Raises:
            RawMemberError: If a member inside `rval` is not defined in the
                schema.
            RawTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        raise NotImplementedError

    def clear_val_counters(self) -> None:
        """Clear receiver's validation counter."""
        self.val_count = 0

    def _apply_deviate(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Apply **deviate** statement to the receiver."""
        arg = stmt.argument
        if arg == "not-supported":
            self.delete()
        else:
            for subst in stmt.substatements:
                try:
                    method = getattr(self, "_deviate_" +
                                     subst.keyword.replace("-", "_", 1))
                except AttributeError as error:
                    if subst.prefix:  # ignore unsupported extension
                        continue
                    raise error from None
                method(subst, sctx, action=arg)

    def _get_description(self, stmt: Statement):
        dst = stmt.find1("description")
        if dst is not None:
            self.description = dst.argument

    def _yang_class(self) -> str:
        return self.__class__.__name__[:-4].lower()

    def _node_digest(self) -> dict[str, Any]:
        """Return dictionary of receiver's properties suitable for clients."""
        res = {"kind": self._yang_class()}
        if self.mandatory:
            res["mandatory"] = True
        if self.description:
            res["description"] = self.description
        return res

    def _tree_name(self) -> str:
        """Return the receiver's name to be displayed in ASCII tree."""
        return (self.name if self.parent and self.ns == self.parent.ns
                else f"{self.ns}:{self.name}")

    def _validate(self, inst: InstanceNode, scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Validate instance against the receiver.

        Args:
            inst: Instance node to be validated.
            scope: Scope of the validation (syntax, semantics or all)
            ctype: Content type of the instance.

        Returns:
            ``None`` if validation succeeds.

        Raises:
            SchemaError: if `inst` doesn't conform to the schema.
            SemanticError: If `inst` violates a semantic rule.
            YangTypeError: If `inst` is a scalar of incorrect type.
        """
        self.val_count += 1

    def _iname2qname(self, iname: InstanceName) -> QualName:
        """Translate instance name to qualified name in the receiver's context.
        """
        p, s, loc = iname.partition(":")
        return (loc, p) if s else (p, self.ns)

    def _flatten(self) -> list["SchemaNode"]:
        return [self]

    def _handle_substatements(self, stmt: Statement,
                              sctx: SchemaContext) -> None:
        """Dispatch actions for substatements of `stmt`."""
        for s in stmt.substatements:
            if s.prefix:
                key = (sctx.schema_data.modules[
                    sctx.text_mid].prefix_map[s.prefix][0]
                    + ":" + s.keyword)
            else:
                key = s.keyword
            mname = SchemaNode._stmt_callback.get(key, "_noop")
            method = getattr(self, mname)
            method(s, sctx)

    def _follow_leafref(self: "SchemaNode", xpath: Expr,
                        init: "TerminalNode") -> Optional["LeafNode"]:
        """Return the data node referred to by a leafref path.

        Args:
            xpath: XPath expression compiled from a leafref path.
            init: initial context node
        """
        if isinstance(xpath, LocationPath):
            lft = self._follow_leafref(xpath.left, init)
            if lft is None:
                return None
            return lft._follow_leafref(xpath.right, init)
        elif isinstance(xpath, Step):
            if xpath.axis == Axis.parent:
                if isinstance(self, SchemaTreeNode):
                    return None
                return self.data_parent() or self.schema_root()
            elif xpath.axis == Axis.child:
                if isinstance(self, InternalNode) and xpath.qname:
                    qname = (xpath.qname if xpath.qname[1]
                             else (xpath.qname[0], init.ns))
                    return self.get_data_child(*qname)
        elif isinstance(xpath, Root):
            return self.schema_root()
        return None

    def _noop(self, stmt: Statement, sctx: SchemaContext) -> None:
        pass

    def _config_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        if stmt.argument == "false":
            self._ctype = ContentType.nonconfig

    def _status_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        if stmt.argument == "deprecated":
            nst = NodeStatus.deprecated
        elif stmt.argument == "obsolete":
            nst = NodeStatus.obsolete
        else:
            nst = NodeStatus.current
        if nst != self.parent.status:
            self._status = nst

    def _deviate_config(self, stmt: Statement,
                        sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self._ctype = None
        elif action in ("add", "replace"):
            self._config_stmt(stmt, sctx)

    def _description_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.description = stmt.argument

    def _must_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        xpp = XPathParser(stmt.argument, sctx)
        mex = xpp.parse()
        if not xpp.at_end():
            raise InvalidArgument(stmt.argument)
        self.must.append(Must(mex, *stmt.get_error_info()))

    def _deviate_must(self, stmt: Statement,
                      sctx: SchemaContext, action: str) -> None:
        if action in ("add", "replace"):
            if action == "replace":
                self.must = []
            self._must_stmt(stmt, sctx)
        elif action == "delete":
            mstr = str(XPathParser(stmt.argument, sctx).parse())
            for i in range(len(self.must)):
                if str(self.must[i].expression) == mstr:
                    del self.must[i]
                    return

    def _when_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        xpp = XPathParser(stmt.argument, sctx)
        wex = xpp.parse()
        if not xpp.at_end():
            raise InvalidArgument(stmt.argument)
        self.when = wex

    def _mandatory_stmt(self, stmt, sctx: SchemaContext) -> None:
        if stmt.argument == "true":
            self._mandatory = True
        elif stmt.argument == "false":
            self._mandatory = False

    def _deviate_mandatory(self, stmt: Statement,
                           sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self._mandatory = False
        elif action in ("add", "replace"):
            self._mandatory_stmt(stmt, sctx)

    def _post_process(self) -> None:
        pass

    def _is_identityref(self) -> bool:
        return False

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return f"{self._tree_line_prefix()} {self._tree_name()}"

    def _tree_line_prefix(self) -> str:
        return self.status.value + "--"

    def _nacm_default_deny_stmt(self, stmt: Statement,
                                sctx: SchemaContext) -> None:
        """Set NACM default access."""
        if not hasattr(self, 'default_deny'):
            return
        if stmt.keyword == "default-deny-all":
            self.default_deny = DefaultDeny.all
        elif stmt.keyword == "default-deny-write":
            self.default_deny = DefaultDeny.write

    _stmt_callback = {
        "action": "_rpc_action_stmt",
        "anydata": "_anydata_stmt",
        "anyxml": "_anydata_stmt",
        "case": "_case_stmt",
        "choice": "_choice_stmt",
        "config": "_config_stmt",
        "container": "_container_stmt",
        "default": "_default_stmt",
        "identity": "_identity_stmt",
        "ietf-netconf-acm:default-deny-all": "_nacm_default_deny_stmt",
        "ietf-netconf-acm:default-deny-write": "_nacm_default_deny_stmt",
        "ietf-yang-metadata:annotation": "_annotation_stmt",
        "input": "_input_stmt",
        "key": "_key_stmt",
        "leaf": "_leaf_stmt",
        "leaf-list": "_leaf_list_stmt",
        "list": "_list_stmt",
        "mandatory": "_mandatory_stmt",
        "max-elements": "_max_elements_stmt",
        "min-elements": "_min_elements_stmt",
        "must": "_must_stmt",
        "notification": "_notification_stmt",
        "output": "_output_stmt",
        "ordered-by": "_ordered_by_stmt",
        "presence": "_presence_stmt",
        "rpc": "_rpc_action_stmt",
        "status": "_status_stmt",
        "unique": "_unique_stmt",
        "units": "_units_stmt",
        "uses": "_uses_stmt",
        "when": "_when_stmt",
    }
    """Map of statement keywords to callback methods."""


class InternalNode(SchemaNode):
    """Abstract class for schema nodes that have children."""

    _mandatory_children: tuple[MutableSet[SchemaNode], MutableSet[SchemaNode]]
    """Two sets of mandatory children: item 0 = nonconfig, 1 = config."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.children: list[SchemaNode] = []
        self.schema_pattern: Optional[SchemaPattern] = None
        self._mandatory_children = (set(), set())

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return bool(self._mandatory_children[0] or
                    self._mandatory_children[1])

    def get_child(self, name: YangIdentifier,
                  ns: YangIdentifier = None) -> Optional[SchemaNode]:
        """Return receiver's schema child.

        Args:
            name: Child's name.
            ns: Child's namespace (= `self.ns` if absent).
        """
        ns = ns if ns else self.ns
        todo = []
        for child in self.children:
            if child.name is None:
                todo.append(child)
            elif child.name == name and child.ns == ns:
                return child
        for c in todo:
            grandchild = c.get_child(name, ns)
            if grandchild is not None:
                return grandchild

    def get_schema_descendant(
            self: "InternalNode", route: SchemaRoute) -> Optional[SchemaNode]:
        """Return descendant schema node or ``None`` if not found.

        Args:
            route: Schema route to the descendant node
                   (relative to the receiver).
        """
        node = self
        for p in route:
            try:
                node = node.get_child(*p)
            except AttributeError:
                return None
            if node is None:
                return None
        return node

    def get_data_child(self, name: YangIdentifier,
                       ns: YangIdentifier = None) -> Optional["DataNode"]:
        """Return data node directly under the receiver."""
        ns = ns if ns else self.ns
        todo = []
        for child in self.children:
            if child.name == name and child.ns == ns:
                if isinstance(child, DataNode):
                    return child
                todo.insert(0, child)
            elif not isinstance(child, DataNode):
                todo.append(child)
        for c in todo:
            res = c.get_data_child(name, ns)
            if res:
                return res

    def filter_children(self, ctype: ContentType = None) -> list[SchemaNode]:
        """Return receiver's children based on content type.

        Args:
            ctype: Content type.
        """
        if ctype is None:
            ctype = self.content_type()
        return [c for c in self.children if
                not isinstance(c, (RpcActionNode, NotificationNode)) and
                c.content_type().value & ctype.value != 0]

    def data_children(self) -> list["DataNode"]:
        """Return the set of all data nodes directly under the receiver."""
        res = []
        for child in self.children:
            if isinstance(child, DataNode):
                res.append(child)
            elif not isinstance(child, SchemaTreeNode):
                res.extend(child.data_children())
        return res

    def from_raw(self, rval: RawObject, jptr: JSONPointer = "") -> ObjectValue:
        """Override the superclass method."""
        if not isinstance(rval, dict):
            raise RawTypeError(jptr, "object")
        res = ObjectValue()
        for qn in rval:
            if qn.startswith("@"):
                if qn != "@":
                    tgt = qn[1:]
                    if tgt not in rval:
                        raise MissingAnnotationTarget(jptr, tgt)
                    jptr += '/' + tgt
                res[qn] = self._process_metadata(rval[qn], jptr)
            else:
                cn = self._iname2qname(qn)
                ch = self.get_data_child(*cn)
                if jptr == "" and ch is None:
                    ch = self.get_child(*cn)
                    if not isinstance(ch, SchemaTreeNode):
                        ch = None
                npath = jptr + "/" + qn
                if ch is None:
                    raise RawMemberError(npath)
                if jptr == "" or self.ns != ch.ns:
                    iname = '{1}:{0}'.format(*ch.qual_name)
                else:
                    iname = ch.name
                res[iname] = ch.from_raw(rval[qn], npath)
        return res

    def from_xml(self, rval: ET.Element,
                 jptr: JSONPointer = "") -> ObjectValue:
        res = ObjectValue()
        if isinstance(self, RpcActionNode) and jptr == "":
            self._process_xmlobj_child(res, None, rval, jptr)
        else:
            for xmlchild in rval:
                self._process_xmlobj_child(res, rval, xmlchild, jptr)
        return res

    def _process_xmlobj_child(
            self: "InternalNode", res: ObjectValue, rval: ET.Element,
            xmlchild: ET.Element, jptr: JSONPointer):
        if xmlchild.tag[0] == '{':
            xmlns, name = xmlchild.tag[1:].split('}')
            nsmap = self.schema_root().schema_data.modules_by_ns
            if xmlns not in nsmap:
                raise MissingModuleNamespace(xmlns)
            ns = nsmap[xmlns].yang_id[0]
            fqn = ns + ':' + name
        else:
            name = xmlchild.tag
            ns = self.ns
            fqn = ns + ':' + name
        qn = fqn if ns != self.ns else name

        ch = self.get_data_child(name, ns)
        if jptr == "" and ch is None:
            ch = self.get_child(name, ns)
            if not isinstance(ch, SchemaTreeNode):
                ch = None
        npath = jptr + "/" + qn
        if ch is None:
            raise RawMemberError(npath)

        if rval is not None and self.ns == ch.ns:
            iname = ch.name
        else:
            iname = '{1}:{0}'.format(*ch.qual_name)
        if isinstance(ch, SequenceNode):
            if iname not in res:
                res[iname] = ch.from_xml(rval, npath, fqn)
        else:
            res[iname] = ch.from_xml(xmlchild, npath)

    def _process_metadata(self, rmo: RawMetadataObject,
                          jptr: JSONPointer) -> MetadataObject:
        res = {}
        ans = self.schema_root().annotations
        for mem in rmo:
            try:
                an = ans[self._iname2qname(mem)]
            except KeyError:
                raise UndefinedAnnotation(jptr, mem)
            res[mem] = an.type.from_raw(rmo[mem])
            if res[mem] not in an.type:
                raise AnnotationTypeError(jptr, mem, an.type.error_message)
        return res

    def _node_digest(self) -> dict[str, Any]:
        res = super()._node_digest()
        rc = res["children"] = {}
        for c in self.data_children():
            cdig = rc[c.iname()] = c._node_digest()
            if self.config and not c.config:
                cdig["config"] = False
        for c in [c for c in self.children if isinstance(c, SchemaTreeNode)]:
            rc[c.iname()] = c._node_digest()
        return res

    def _validate(self, inst: InstanceNode,
                  scope: ValidationScope, ctype: ContentType) -> None:
        """Extend the superclass method."""
        if scope.value & ValidationScope.syntax.value:   # schema
            self._check_schema_pattern(inst, ctype)
        for m in inst:
            inst._member(m).validate(scope, ctype)
        super()._validate(inst, scope, ctype)

    def _add_child(self, node: SchemaNode) -> None:
        node.parent = self
        self.children.append(node)

    def _child_inst_names(self) -> set[InstanceName]:
        """Return the set of instance names under the receiver."""
        return frozenset([c.iname() for c in self.data_children()])

    def _check_schema_pattern(self, inst: InstanceNode,
                              ctype: ContentType) -> None:
        p = self.schema_pattern
        p._eval_when(inst)
        for m in inst:
            newp = p.deriv(m, ctype)
            if isinstance(newp, NotAllowed):
                raise SchemaError(
                    inst,
                    ("" if ctype == ContentType.all else ctype.name + " ") +
                    "member-not-allowed", m)
            p = newp
        if not p.nullable(ctype):
            mms = p._mandatory_members(ctype)
            msg = "one of " if len(mms) > 1 else ""
            raise SchemaError(
                inst, "missing-data",
                "expected " + msg + ", ".join([repr(m) for m in mms]))

    def _make_schema_patterns(self) -> None:
        """Build schema pattern for the receiver and its data descendants."""
        self.schema_pattern = self._schema_pattern()
        for dc in self.data_children():
            if isinstance(dc, InternalNode):
                dc._make_schema_patterns()

    def _schema_pattern(self) -> SchemaPattern:
        todo = [c for c in self.children if not (
            isinstance(c, (RpcActionNode, NotificationNode)) or
            c._status == NodeStatus.obsolete)]
        if not todo:
            return Empty()
        prev = todo[0]._pattern_entry()
        for c in todo[1:]:
            prev = Pair(c._pattern_entry(), prev)
        return ConditionalPattern(prev, self.when) if self.when else prev

    def _post_process(self) -> None:
        super()._post_process()
        for c in self.children:
            c._post_process()

    def _add_mandatory_child(self, node: SchemaNode) -> None:
        """Add `node` to the set of mandatory children."""
        if node.mandatory_config:
            self._mandatory_children[1].add(node)
        else:
            self._mandatory_children[0].add(node)

    def _add_defaults(self, inst: InstanceNode, ctype: Optional[ContentType],
                      lazy: bool = False) -> InstanceNode:
        for c in self.filter_children(ctype):
            if isinstance(c, DataNode):
                inst = c._default_instance(inst, ctype, lazy)
            elif not isinstance(c, (RpcActionNode, NotificationNode)):
                inst = c._add_defaults(inst, ctype)
        return inst

    def _state_roots(self) -> list[SchemaNode]:
        if self.content_type() == ContentType.nonconfig:
            return [self]
        res = []
        for c in self.data_children():
            res.extend(c._state_roots())
        return res

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                      sctx: SchemaContext) -> None:
        """Add child node to the receiver and handle substatements."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        node.name = stmt.argument
        node.ns = sctx.default_ns
        node._get_description(stmt)
        self._add_child(node)
        node._handle_substatements(stmt, sctx)

    def _augment_stmt(self, stmt: Statement,
                      sctx: SchemaContext) -> None:
        """Handle **augment** statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        target = self.get_schema_descendant(
            sctx.schema_data.sni2route(stmt.argument, sctx))
        if target is None:      # silently ignore missing target
            return
        if stmt.find1("when"):
            gr = GroupNode()
            target._add_child(gr)
            target = gr
        target._handle_substatements(stmt, sctx)

    def _deviation_stmt(self, stmt: Statement,
                        sctx: SchemaContext) -> None:
        """Handle **deviation** statement."""
        target = self.get_schema_descendant(
            sctx.schema_data.sni2route(stmt.argument, sctx))
        if target is None:      # silently ignore missing target
            return
        for dstmt in stmt.find_all("deviate"):
            target._apply_deviate(dstmt, sctx)

    def _refine_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle **refine** statement."""
        target = self.get_schema_descendant(
            sctx.schema_data.sni2route(stmt.argument, sctx))
        if not target:
            return
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            target.parent.children.remove(target)
        else:
            target._handle_substatements(stmt, sctx)

    def _uses_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle uses statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        grp, gid = sctx.schema_data.get_definition(stmt, sctx)
        wst = stmt.find1("when")
        if wst:
            sn = GroupNode()
            xpp = XPathParser(wst.argument, sctx)
            wex = xpp.parse()
            if not xpp.at_end():
                raise InvalidArgument(wst.argument)
            sn.when = wex
            self._add_child(sn)
        else:
            sn = self
        sn._handle_substatements(grp, gid)
        for augst in stmt.find_all("augment"):
            sn._augment_stmt(augst, sctx)
        for refst in stmt.find_all("refine"):
            sn._refine_stmt(refst, sctx)

    def _container_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle container statement."""
        self._handle_child(ContainerNode(), stmt, sctx)

    def _identity_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle identity statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        id = (stmt.argument, sctx.schema_data.namespace(sctx.text_mid))
        adj = sctx.schema_data.identity_adjs.setdefault(
            id, IdentityAdjacency())
        for bst in stmt.find_all("base"):
            bid = sctx.schema_data.translate_pname(bst.argument, sctx.text_mid)
            adj.bases.add(bid)
            badj = sctx.schema_data.identity_adjs.setdefault(
                bid, IdentityAdjacency())
            badj.derivs.add(id)
        sctx.schema_data.identity_adjs[id] = adj

    def _list_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle list statement."""
        self._handle_child(ListNode(), stmt, sctx)

    def _choice_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle choice statement."""
        self._handle_child(ChoiceNode(), stmt, sctx)

    def _case_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle case statement."""
        self._handle_child(CaseNode(), stmt, sctx)

    def _leaf_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle leaf statement."""
        node = LeafNode()
        node.type = DataType._resolve_type(
            stmt.find1("type", required=True), sctx)
        self._handle_child(node, stmt, sctx)

    def _leaf_list_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle leaf-list statement."""
        node = LeafListNode()
        node.type = DataType._resolve_type(
            stmt.find1("type", required=True), sctx)
        self._handle_child(node, stmt, sctx)

    def _rpc_action_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle rpc or action statement."""
        self._handle_child(RpcActionNode(), stmt, sctx)

    def _notification_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle notification statement."""
        self._handle_child(NotificationNode(), stmt, sctx)

    def _anydata_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle anydata statement."""
        self._handle_child(AnydataNode(), stmt, sctx)

    def _ascii_tree(self, indent: str, no_types: bool, val_count: bool) -> str:
        """Return the receiver's subtree as ASCII art."""
        def suffix(sn):
            return f" {{{sn.val_count}}}\n" if val_count and isinstance(
                sn, (SchemaTreeNode, DataNode)) else "\n"
        if not self.children:
            return ""
        cs = []
        for c in self.children:
            cs.extend(c._flatten())
        cs.sort(key=lambda x: x.qual_name)
        res = ""
        for c in cs[:-1]:
            res += (indent + c._tree_line(no_types) + suffix(c) +
                    c._ascii_tree(indent + "|  ", no_types, val_count))
        return (res + indent + cs[-1]._tree_line(no_types) + suffix(cs[-1]) +
                cs[-1]._ascii_tree(indent + "   ", no_types, val_count))

    def clear_val_counters(self) -> None:
        """Clear validation counters in the receiver and its subtree."""
        super().clear_val_counters()
        for c in self.children:
            c.clear_val_counters()


class GroupNode(InternalNode):
    """Anonymous group of schema nodes."""

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                      sctx: SchemaContext) -> None:
        if not isinstance(
                self.parent, ChoiceNode) or isinstance(node, CaseNode):
            super()._handle_child(node, stmt, sctx)
        else:
            cn = CaseNode()
            cn.name = stmt.argument
            cn.ns = sctx.default_ns
            self._add_child(cn)
            cn._handle_child(node, stmt, sctx)

    def _pattern_entry(self) -> SchemaPattern:
        return super()._schema_pattern()

    def _flatten(self) -> list[SchemaNode]:
        res = []
        for c in self.children:
            res.extend(c._flatten())
        return res


class SchemaTreeNode(GroupNode):
    """Root node of a schema tree."""

    def __init__(self, schemadata: SchemaData = None):
        """Initialize the class instance."""
        super().__init__()
        self.annotations: dict[QualName, Annotation] = {}
        self.schema_data = schemadata
        self._status = NodeStatus.current

    def iname(self) -> InstanceName:
        """Override the superclass method."""
        return "/"

    def data_parent(self) -> None:
        """Override the superclass method."""
        return None

    def _annotation_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle annotation statement."""
        if not sctx.schema_data.if_features(stmt, sctx.text_mid):
            return
        dst = stmt.find1("description")
        self.annotations[(stmt.argument, sctx.default_ns)] = Annotation(
            DataType._resolve_type(stmt.find1("type", required=True), sctx),
            dst.argument if dst else None)


class DataNode(SchemaNode):
    """Abstract superclass for all data nodes."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny: DefaultDeny = DefaultDeny.none

    def orphan_instance(self, rval: RawValue) -> "ObjectMember":
        """Return an isolated instance of the receiver.

        Args:
            rval: Raw value to be used for the returned instance.
        """
        val = self.from_raw(rval, jptr="/")
        return ObjectMember(self.iname(), {}, val, None, self, datetime.now())

    def split_instance_route(self, route: InstanceRoute) -> Optional[
            tuple[InstanceRoute, InstanceRoute]]:
        """Split `route` into the part up to receiver and the rest.

        Args:
            route: Absolute instance route (the receiver should correspond
                to an instance node on this route).

        Returns:
            A tuple consisting of
                - the part of `route` from the root up to and including the
                  instance whose schema node is the receiver, and
                - the rest of `route`.
                ``None`` is returned if the receiver is not on the route.
        """
        sroute = []
        sn = self
        while sn:
            sroute.append(sn.iname())
            sn = sn.data_parent()
        i = 0
        while True:
            if not sroute:
                break
            inst = sroute.pop()
            if inst != route[i].iname():
                return None
            while True:         # skip up to next MemberName
                i += 1
                if i >= len(route) or isinstance(route[i], MemberName):
                    break
            if not sroute:
                return (InstanceRoute(route[:i]), InstanceRoute(route[i:]))
            if i >= len(route):
                return None

    def _validate(self, inst: InstanceNode, scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Extend the superclass method."""
        if scope.value & ValidationScope.semantics.value:
            self._check_must(inst)        # must expressions
        super()._validate(inst, scope, ctype)

    def _default_instance(self, pnode: InstanceNode, ctype: ContentType,
                          lazy: bool = False) -> InstanceNode:
        iname = self.iname()
        if iname in pnode.value:
            return pnode
        nm = pnode.put_member(iname, (None,))
        if not self.when or self.when.evaluate(nm):
            wd = self._default_value(nm, ctype, lazy)
            if wd.value is not None:
                return wd.up()
        return pnode

    def _check_must(self, inst: InstanceNode) -> None:
        for m in self.must:
            if not m.expression.evaluate(inst):
                raise SemanticError(inst, m.error_tag, m.error_message)

    def _pattern_entry(self) -> SchemaPattern:
        m = Member(self.iname(), self.content_type(), self.when)
        return m if (self.mandatory and self._status !=
                     NodeStatus.deprecated) else SchemaPattern.optional(m)

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + (
            "ro" if self.content_type() == ContentType.nonconfig else "rw")


class TerminalNode(SchemaNode):
    """Abstract superclass for terminal nodes in the schema tree."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.type: DataType = None
        self._default: Optional[Value] = None
        self._units: Optional[str] = None

    def content_type(self) -> ContentType:
        """Override superclass method."""
        if self._ctype:
            return self._ctype
        return (ContentType.config if self.parent.config else
                ContentType.nonconfig)

    @property
    def units(self) -> Optional[str]:
        """Units of the receiver's value, if specified."""
        return (self._units if self._units is not None
                else self.type.units)

    def from_raw(self, rval: RawScalar,
                 jptr: JSONPointer = "") -> ScalarValue:
        """Override the superclass method."""
        res = self.type.from_raw(rval)
        if res is None:
            raise RawTypeError(jptr, self.type.yang_type() + " value")
        return res

    def from_xml(self, rval: ET.Element, jptr: JSONPointer = "") -> Value:
        res = self.type.from_xml(rval)
        if res is None:
            raise RawTypeError(jptr, self.type.yang_type() + " value")
        return res

    def _deviate_type(self, stmt: Statement,
                      sctx: SchemaContext, action: str) -> None:
        if action == "replace":
            self.type = DataType._resolve_type(stmt, sctx)

    def _node_digest(self) -> dict[str, Any]:
        res = super()._node_digest()
        res["type"] = self.type._type_digest(self.config)
        df = self.default
        if df is not None:
            res["default"] = self.type.to_raw(df)
        if self.units:
            res["units"] = self.units
        return res

    def _units_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self._units = stmt.argument

    def _deviate_units(self, stmt: Statement,
                       sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self._units = None
        elif action in ("add", "replace"):
            self._units_stmt(stmt, sctx)

    def _validate(self, inst: InstanceNode,
                  scope: ValidationScope, ctype: ContentType) -> None:
        """Extend the superclass method."""
        if (scope.value & ValidationScope.syntax.value and
                inst.value not in self.type):
            raise YangTypeError(inst, self.type.error_tag,
                                self.type.error_message)
        if (isinstance(self.type, LinkType) and        # referential integrity
                scope.value & ValidationScope.semantics.value and
                self.type.require_instance):
            try:
                tgt = inst._deref()
            except YangsonException:
                tgt = []
            if not tgt:
                raise SemanticError(inst, "instance-required")
        super()._validate(inst, scope, ctype)

    def _default_value(self, inst: InstanceNode, ctype: ContentType,
                       lazy: bool) -> InstanceNode:
        inst.value = self.default
        return inst

    def _post_process(self) -> None:
        super()._post_process()
        self.type._post_process(self)

    def _is_identityref(self) -> bool:
        return isinstance(self.type, IdentityrefType)

    def _ascii_tree(self, indent: str, no_types: bool, val_count: bool) -> str:
        return ""

    def _state_roots(self) -> list[SchemaNode]:
        return [] if self.content_type() == ContentType.config else [self]


class ContainerNode(DataNode, InternalNode):
    """Container node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.presence: bool = False

    @property
    def mandatory(self) -> bool:
        """Extend the superclass property."""
        return not self.presence and super().mandatory

    @property
    def mandatory_config(self) -> bool:
        """Override the superclass property."""
        return bool(self._mandatory_children[1])

    def _node_digest(self) -> dict[str, Any]:
        res = super()._node_digest()
        res["presence"] = self.presence
        return res

    def _add_mandatory_child(self, node: SchemaNode):
        propagate = not (self.presence or self.mandatory)
        super()._add_mandatory_child(node)
        if propagate:
            self.parent._add_mandatory_child(self)

    def _default_instance(self, pnode: InstanceNode, ctype: ContentType,
                          lazy: bool = False) -> InstanceNode:
        if self.presence:
            return pnode
        return super()._default_instance(pnode, ctype, lazy)

    def _default_value(self, inst: InstanceNode, ctype: ContentType,
                       lazy: bool) -> Optional[InstanceNode]:
        inst.value = ObjectValue()
        return inst if lazy else self._add_defaults(inst, ctype)

    def _pattern_entry(self) -> SchemaPattern:
        m = Member(self.iname(), self.content_type(), self.when)
        if self._status == NodeStatus.deprecated:
            return SchemaPattern.optional(m)
        if self.mandatory:
            return (m if self.mandatory_config else
                    SchemaPattern.optional_config(m))
        else:
            return SchemaPattern.optional(m)

    def _presence_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.presence = True

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return super()._tree_line() + ("!" if self.presence else "")


class SequenceNode(DataNode):
    """Abstract class for data nodes that represent a sequence."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.min_elements: int = 0
        self.max_elements: Optional[int] = None
        self.user_ordered: bool = False

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return self.min_elements > 0

    def _validate(self, inst: InstanceNode,
                  scope: ValidationScope,
                  ctype: ContentType) -> None:
        """Extend the superclass method."""
        if isinstance(inst, ArrayEntry):
            super()._validate(inst, scope, ctype)
        else:
            if scope.value & ValidationScope.semantics.value:
                self._check_list_props(inst)
                self._check_cardinality(inst)
            for e in inst:
                super()._validate(e, scope, ctype)

    def _check_cardinality(self, inst: InstanceNode) -> None:
        if len(inst.value) < self.min_elements:
            raise SemanticError(inst, "too-few-elements")
        if (self.max_elements is not None and
                len(inst.value) > self.max_elements):
            raise SemanticError(inst, "too-many-elements")

    def _post_process(self) -> None:
        super()._post_process()
        if self.min_elements > 0:
            self.parent._add_mandatory_child(self)

    def _min_elements_stmt(self, stmt: Statement,
                           sctx: SchemaContext) -> None:
        self.min_elements = int(stmt.argument)

    def _deviate_min_elements(self, stmt: Statement,
                              sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self.min_elements = 0
        elif action in ("add", "replace"):
            self._min_elements_stmt(stmt, sctx)

    def _max_elements_stmt(self, stmt: Statement,
                           sctx: SchemaContext) -> None:
        arg = stmt.argument
        if arg == "unbounded":
            self.max_elements = None
        else:
            self.max_elements = int(arg)

    def _deviate_max_elements(self, stmt: Statement,
                              sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self.max_elements = None
        elif action in ("add", "replace"):
            self._max_elements_stmt(stmt, sctx)

    def _ordered_by_stmt(self, stmt: Statement,
                         sctx: SchemaContext) -> None:
        self.user_ordered = stmt.argument == "user"

    def _tree_line(self, no_type: bool = False) -> str:
        """Extend the superclass method."""
        return super()._tree_line() + ("#" if self.user_ordered else "*")

    def from_raw(self, rval: RawList,
                 jptr: JSONPointer = "") -> ArrayValue:
        """Override the superclass method."""
        if not isinstance(rval, list):
            raise RawTypeError(jptr, "array")
        res = ArrayValue()
        idx = 0
        for i in range(len(rval)):
            if isinstance(self, ListNode):
                try:
                    keys = [str(rval[i].get(k[0], '<missing>'))
                            for k in self.keys]
                except AttributeError:
                    raise RawTypeError(f"{jptr}/{i}", "object")
                element = "=" + ",".join(keys)
            else:
                element = "/" + str(idx)
            res.append(self.entry_from_raw(rval[i], jptr + element))
            idx = idx + 1
        return res

    def from_xml(self, rval: ET.Element, jptr: JSONPointer = "",
                 tagname: str = None, isroot: bool = False) -> ArrayValue:
        res = ArrayValue()
        idx = 0
        if isroot:
            return self._process_xmlarray_child(res, rval, None, jptr)
        else:
            for xmlchild in rval:
                if isinstance(self, ListNode):
                    keys = [str(xmlchild.findtext(
                        k[0], default='<missing>')) for k in self.keys]
                    element = "=" + ",".join(keys)
                else:
                    element = "/" + str(idx)
                self._process_xmlarray_child(
                    res, xmlchild, tagname, jptr + element)
                idx = idx + 1
        return res

    def _process_xmlarray_child(
            self: "SequenceNode", res: ArrayValue, xmlchild: ET.Element,
            tagname: str, jptr: JSONPointer):
        if xmlchild.tag[0] == '{':
            xmlns, name = xmlchild.tag[1:].split('}')
            module = self.schema_root().schema_data.modules_by_ns.get(xmlns)
            if not module:
                raise MissingModuleNamespace(xmlns)
            ns = module.yang_id[0]
            qn = ns + ':' + name
        else:
            name = qn = xmlchild.tag
            ns = self.ns
        if tagname is None or qn == tagname:
            child = self.entry_from_xml(xmlchild, jptr + "/" + str(len(res)))
            res.append(child)
        else:
            child = None
        return child

    def entry_from_raw(self, rval: RawEntry,
                       jptr: JSONPointer = "") -> EntryValue:
        """Transform a raw (leaf-)list entry into the cooked form.

        Args:
            rval: raw entry (scalar or object)
            jptr: JSON pointer of the entry

        Raises:
            NonexistentSchemaNode: If a member inside `rval` is not defined
                in the schema.
            RawTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        return super().from_raw(rval, jptr)

    def entry_from_xml(self, rval: ET.Element,
                       jptr: JSONPointer = "") -> EntryValue:
        """Transform a XML (leaf-)list entry into the cooked form.

        Args:
            rval: xml node
            jptr: JSON pointer of the entry

        Raises:
            NonexistentSchemaNode: If a member inside `rval` is not defined
                in the schema.
            RawTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        return super().from_xml(rval, jptr)


class ListNode(SequenceNode, InternalNode):
    """List node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.keys: list[QualName] = []
        self._key_members = []
        self.unique: list[list[LocationPath]] = []

    def _node_digest(self) -> dict[str, Any]:
        res = super()._node_digest()
        res["keys"] = self._key_members
        return res

    def _check_list_props(self, inst: InstanceNode) -> None:
        """Check uniqueness of keys and "unique" properties, if applicable."""
        if self.keys:
            self._check_keys(inst)
        for u in self.unique:
            self._check_unique(u, inst)

    def _check_keys(self, inst: InstanceNode) -> None:
        ukeys = set()
        for i in range(len(inst.value)):
            en = inst.value[i]
            try:
                kval = tuple([en[k] for k in self._key_members])
            except KeyError as e:
                raise SchemaError(inst._entry(i),
                                  "list-key-missing", e.args[0]) from None
            if kval in ukeys:
                raise SemanticError(inst, "non-unique-key",
                                    repr(kval[0] if len(kval) < 2 else kval))
            ukeys.add(kval)

    def _check_unique(self, unique: list[LocationPath],
                      inst: InstanceNode) -> None:
        allvals = set()
        for i in range(len(inst.value)):
            en = inst[i]
            den = en.add_defaults()
            uvals = []
            for uex in unique:
                uval = [n.value for n in uex.evaluate(den)]
                uvals.append(uval)
            tups = set(product(*uvals))
            if tups & allvals:
                raise SemanticError(inst, f"data-not-unique: entry {i}")
            else:
                allvals |= tups

    def _default_instance(self, pnode: InstanceNode, ctype: ContentType,
                          lazy: bool = False) -> InstanceNode:
        return pnode

    def _post_process(self) -> None:
        super()._post_process()
        for k in self.keys:
            kn = self.get_data_child(*k)
            self._key_members.append(kn.iname())
            if not kn._mandatory:
                kn._mandatory = True
                self._add_mandatory_child(kn)

    def _key_stmt(self, stmt: Statement,
                  sctx: SchemaContext) -> None:
        self.keys = [sctx.schema_data.translate_node_id(k, sctx)
                     for k in stmt.argument.split()]

    def _parse_unique(self, stmt: Statement,
                      sctx: SchemaContext) -> list[LocationPath]:
        uspec = []
        for sid in stmt.argument.split():
            xpp = XPathParser(sid, sctx)
            uex = xpp.parse()
            if not xpp.at_end():
                raise InvalidArgument(stmt.argument)
            uspec.append(uex)
        return uspec

    def _unique_stmt(self, stmt: Statement,
                     sctx: SchemaContext) -> None:
        self.unique.append(self._parse_unique(stmt, sctx))

    def _deviate_unique(self, stmt: Statement,
                        sctx: SchemaContext, action: str) -> None:
        if action in ("add", "replace"):
            if action == "replace":
                self.unique = []
            self._unique_stmt(stmt, sctx)
        elif action == "delete":
            uniq = self._parse_unique(stmt, sctx)
            for i in range(len(self.unique)):
                if self.unique[i] == uniq:
                    del self.unique[i]
                    return

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        keys = (" [" + " ".join([k[0] for k in self.keys]) + "]"
                if self.keys else "")
        return super()._tree_line() + keys

    def orphan_entry(self, rval: RawObject) -> "ArrayEntry":
        """Return an isolated entry of the receiver.

        Args:
            rval: Raw object to be used for the returned entry.
        """
        val = self.entry_from_raw(rval, jptr="/")
        return ArrayEntry(0, deque(), deque(), val, None, self,
                          val.timestamp)


class ChoiceNode(InternalNode):
    """Choice node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_case: QualName = None
        self._mandatory: bool = False

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return self._mandatory

    def _add_defaults(self, inst: InstanceNode,
                      ctype: Optional[ContentType]) -> InstanceNode:
        if self.when and not self.when.evaluate(inst):
            return inst
        ac = self._active_case(inst.value)
        if ac:
            return ac._add_defaults(inst, ctype)
        elif self.default_case:
            n = dc = self.get_child(*self.default_case)
            while n is not self:
                if n.when and not n.when.evaluate(inst):
                    return inst
                n = n.parent
            return dc._add_defaults(inst, ctype)
        else:
            return inst

    def _active_case(self, value: ObjectValue) -> Optional["CaseNode"]:
        """Return receiver's case that's active in an instance node value."""
        for c in self.children:
            for cc in c.data_children():
                if cc.iname() in value:
                    return c

    def _pattern_entry(self) -> SchemaPattern:
        if not self.children:
            return Empty()
        prev = self.children[0]._schema_pattern()
        for c in self.children[1:]:
            prev = ChoicePattern(c._schema_pattern(), prev, self.name)
        prev.ctype = self.content_type()
        if not self.mandatory:
            prev = SchemaPattern.optional(prev)
        return ConditionalPattern(prev, self.when) if self.when else prev

    def _post_process(self) -> None:
        super()._post_process()
        if self.mandatory:
            self.parent._add_mandatory_child(self)

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + (
            "ro" if self.content_type() == ContentType.nonconfig else "rw")

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                      sctx: SchemaContext) -> None:
        if isinstance(node, CaseNode):
            super()._handle_child(node, stmt, sctx)
        else:
            cn = CaseNode()
            cn.name = stmt.argument
            cn.ns = sctx.default_ns
            self._add_child(cn)
            cn._handle_child(node, stmt, sctx)

    def _default_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self.default_case = sctx.schema_data.translate_node_id(
            stmt.argument, sctx)

    def _deviate_default(self, stmt: Statement,
                         sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self.default_case = None
        elif action in ("add", "replace"):
            self._default_stmt(stmt, sctx)

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return f"{self._tree_line_prefix()} ({self._tree_name()})" \
            f"{'' if self._mandatory else '?'}"


class CaseNode(InternalNode):
    """Case node."""

    def _pattern_entry(self) -> SchemaPattern:
        return super()._schema_pattern()

    def _tree_line(self, no_type: bool = False) -> str:
        """Return the receiver's contribution to tree diagram."""
        return f"{self._tree_line_prefix()}:({self._tree_name()})"


class LeafNode(DataNode, TerminalNode):
    """Leaf node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._mandatory: bool = False

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return self._mandatory

    @property
    def default(self) -> Optional[ScalarValue]:
        """Default value of the receiver, if any."""
        if self.mandatory:
            return None
        if self._default is not None:
            return self._default
        return self.type.default

    def _post_process(self) -> None:
        super()._post_process()
        if self._mandatory:
            self.parent._add_mandatory_child(self)
        elif self._default is not None:
            self._default = self.type.from_yang(self._default)

    def _tree_line(self, no_type: bool = False) -> str:
        res = super()._tree_line() + ("" if self._mandatory else "?")
        return res if no_type else f"{res} <{self.type}>"

    def _default_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        self._default = stmt.argument

    def _deviate_default(self, stmt: Statement,
                         sctx: SchemaContext, action: str) -> None:
        if action == "delete":
            self._default = None
        elif action in ("add", "replace"):
            self._default_stmt(stmt, sctx)


class LeafListNode(SequenceNode, TerminalNode):
    """Leaf-list node."""

    @property
    def default(self) -> Optional[ScalarValue]:
        """Default value of the receiver, if any."""
        if self.mandatory:
            return None
        if self._default is not None:
            return self._default
        return (None if self.type.default is None
                else ArrayValue([self.type.default]))

    def _yang_class(self) -> str:
        return "leaf-list"

    def _check_list_props(self, inst: InstanceNode) -> None:
        if (self.content_type() == ContentType.config and
                len(set(inst.value)) < len(inst.value)):
            raise SemanticError(inst, "repeated-leaf-list-value")

    def _default_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        if self._default is None:
            self._default = [stmt.argument]
        else:
            self._default.append(stmt.argument)

    def _deviate_default(self, stmt: Statement,
                         sctx: SchemaContext, action: str) -> None:
        if action in ("add", "replace"):
            if action == "replace":
                self._default = []
            self._default_stmt(stmt, sctx)
        elif action == "delete":
            val = self.type.parse_value(stmt.argument)
            for i in range(len(self._default)):
                if self._default[i] == val:
                    del self._default[i]
                    return

    def _post_process(self) -> None:
        super()._post_process()
        if self._default is not None:
            self._default = ArrayValue(
                [self.type.from_yang(v) for v in self._default])

    def _tree_line(self, no_type: bool = False) -> str:
        res = super()._tree_line()
        return res if no_type else f"{res} <{self.type}>"


class AnyContentNode(DataNode):
    """Abstract class for anydata or anyxml nodes."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._mandatory: bool = False

    def content_type(self) -> ContentType:
        """Override superclass method."""
        return TerminalNode.content_type(self)

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return self._mandatory

    def from_raw(self, rval: RawValue, jptr: JSONPointer = "") -> Value:
        """Override the superclass method."""
        def convert(val):
            if isinstance(val, list):
                res = ArrayValue([convert(x) for x in val])
            elif isinstance(val, dict):
                res = ObjectValue({x: convert(val[x]) for x in val})
            else:
                res = val
            return res
        return convert(rval)

    def to_raw(self, value: Value) -> RawValue:
        """Convert the value again to plain Python stuff."""
        def convert(val):
            if isinstance(val, ArrayValue):
                res = [convert(x) for x in val]
            elif isinstance(val, ObjectValue):
                res = {x: convert(val[x]) for x in val}
            else:
                res = val
            return res
        return convert(value)

    def from_xml(self, rval: ET.Element, jptr: JSONPointer = "") -> Value:
        super().from_xml(rval, jptr)

    def _default_instance(self, pnode: InstanceNode, ctype: ContentType,
                          lazy: bool = False) -> InstanceNode:
        return pnode

    def _tree_line(self, no_type: bool = False) -> str:
        return super()._tree_line() + ("" if self._mandatory else "?")

    def _ascii_tree(self, indent: str, no_types: bool, val_count: bool) -> str:
        return ""

    def _post_process(self) -> None:
        if self._mandatory:
            self.parent._add_mandatory_child(self)


class AnydataNode(AnyContentNode):
    """Anydata node."""
    pass


class AnyxmlNode(AnyContentNode):
    """Anyxml node."""
    pass


class RpcActionNode(SchemaTreeNode):
    """RPC or action node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny: DefaultDeny = DefaultDeny.none
        self._ctype = ContentType.nonconfig

    def iname(self) -> InstanceName:
        """Override the superclass method."""
        return super(GroupNode, self).iname()

    def data_parent(self) -> None:
        """Override the superclass method."""
        return self.parent

    def _handle_substatements(self, stmt: Statement,
                              sctx: SchemaContext) -> None:
        self._add_child(InputNode(sctx.default_ns))
        self._add_child(OutputNode(sctx.default_ns))
        super()._handle_substatements(stmt, sctx)

    def _flatten(self) -> list[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "-x"

    def _input_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle RPC or action input statement."""
        self.get_child("input")._handle_substatements(stmt, sctx)

    def _output_stmt(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle RPC or action output statement."""
        self.get_child("output")._handle_substatements(stmt, sctx)


class InputNode(InternalNode, DataNode):
    """RPC or action input node."""

    def __init__(self, ns):
        """Initialize the class instance."""
        super().__init__()
        self._config = False
        self.name = "input"
        self.ns = ns

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return False

    def iname(self) -> InstanceName:
        """Override the superclass method."""
        return self.ns + ":" + self.name

    def _flatten(self) -> list[SchemaNode]:
        return [self]


class OutputNode(InternalNode, DataNode):
    """RPC or action output node."""

    def __init__(self, ns):
        """Initialize the class instance."""
        super().__init__()
        self._config = False
        self.name = "output"
        self.ns = ns

    @property
    def mandatory(self) -> bool:
        """Override the superclass property."""
        return False

    def iname(self) -> InstanceName:
        """Override the superclass method."""
        return self.ns + ":" + self.name

    def _flatten(self) -> list[SchemaNode]:
        return [self]


class NotificationNode(SchemaTreeNode):
    """Notification node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny: DefaultDeny = DefaultDeny.none
        self._ctype = ContentType.nonconfig

    def _flatten(self) -> list[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "-n"



# Copyright © 2016–2026 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Classes representing YANG data types.

This module implements the following classes:

* BitsType: YANG bits type.
* BinaryType: YANG binary type.
* BooleanType: YANG boolean type.
* DataType: Abstract class for data types.
* Decimal64Type: YANG decimal64 type.
* EmptyType: YANG empty type.
* EnumerationType: YANG enumeration type.
* IdentityrefType: YANG identityref type.
* InstanceIdentifierType: YANG instance-identifier type.
* LeafrefType: YANG leafref type.
* LinkType: Abstract class for data types representing links.
* IntegralType: Abstract class for integral types.
* Int8Type: YANG int8 type.
* Int16Type: YANG int16 type.
* Int32Type: YANG int32 type.
* Int64Type: YANG int64 type.
* NumericType: Abstract class for numeric types.
* StringType: YANG string type.
* Uint8Type: YANG uint8 type.
* Uint16Type: YANG uint16 type.
* Uint32Type: YANG uint32 type.
* Uint64Type: YANG uint64 type.
* UnionType: YANG union type.
"""
from abc import ABC, abstractmethod
import base64
import binascii
import decimal
import xml.etree.ElementTree as ET
import re
from typing import (Any, cast, ClassVar, Generic, Optional, Type,
                    TYPE_CHECKING)

from .constraint import Intervals, Pattern
from .exceptions import (
    InvalidArgument, ParserException, ModuleNotRegistered, UnknownPrefix,
    InvalidLeafrefPath, MissingModuleNamespace, XPathTypeError)
from .schemadata import SchemaContext
from .instance import InstanceNode, InstanceIdParser
from .instroute import InstanceRoute
from .nodeset import NodeSet
from .statement import Statement
from .typealiases import (InstanceIdentifier, L, N, RN, QualName, RawScalar,
                          RS, S, ScalarValue, YangIdentifier)
from .xpathparser import Expr, XPathParser
if TYPE_CHECKING:
    from .schemanode import TerminalNode


class DataType(ABC, Generic[S, RS]):
    """Abstract class for YANG data types."""

    dtypes: ClassVar[dict[str, Type["DataType"]]]
    _option_template = '<option value="{}"{}>{}</option>'

    def __init__(self, sctx: SchemaContext, name: Optional[YangIdentifier]):
        """Initialize the class instance."""
        self.sctx = sctx
        self.default: Optional[S] = None
        self.name = name
        self.error_tag: Optional[str] = None
        self.error_message: Optional[str] = None
        self.units: Optional[str] = None

    @abstractmethod
    def __contains__(self, val: S) -> bool:
        """Return ``True`` if the receiver type contains `val`.

        If the result is ``False``, set also `error_tag` and `error_message`
        properties.
        """

    def __str__(self):
        """Return YANG name of the receiver type."""
        base = self.yang_type()
        return f"{self.name}({base})" if self.name else base

    @abstractmethod
    def from_raw(self, raw: RS) -> Optional[S]:
        """Return a cooked value of the receiver type.

        The input argument should follow the rules for JSON
        representation of scalar values as specified in
        [RFC7951]_. Conformance to the receiving type isn't
        guaranteed.

        Args:
            raw: Raw value obtained from JSON parser.

        """

    @abstractmethod
    def from_xml(self, xml: ET.Element) -> Optional[S]:
        """Return a cooked value of the received XML type.

        Args:
            xml: Text of the XML node
        """

    @abstractmethod
    def to_raw(self, val: S) -> Optional[RS]:
        """Return a raw value ready to be serialized in JSON."""

    def to_xml(self, val: S) -> Optional[str]:
        """Return XML text value ready to be serialized in XML."""
        return str(val) if val in self else None

    @abstractmethod
    def parse_value(self, text: str) -> Optional[S]:
        """Parse value of the receiver's type.

        The input text should follow the rules for lexical
        representation of scalar values as specified in
        [RFC7950]_. Conformance to the receiving type isn't
        guaranteed.

        Args:
            text: String representation of the value.

        Returns:
            A value of the receiver's type or ``None`` if parsing fails.

        """

    def canonical_string(self, val: S) -> Optional[str]:
        """Return canonical form of a value."""
        return str(val) if val in self else None

    def from_yang(self, text: str) -> S:
        """Parse value specified as default in a YANG module.

        Conformance to the receiving type isn't guaranteed.

        Args:
            text: String representation of the value.

        Raises:
            InvalidArgument: If the receiver type cannot parse the text.
        """
        res = self.parse_value(text)
        if res is None:
            raise InvalidArgument(text)
        return res

    def yang_type(self) -> YangIdentifier:
        """Return YANG name of the receiver."""
        return self.__class__.__name__[:-4].lower()

    def _set_error_info(self, error_tag: Optional[str] = None,
                        error_message: Optional[str] = None):
        self.error_tag = error_tag if error_tag else "invalid-type"
        self.error_message = (error_message if error_message else
                              "expected " + str(self))

    def _post_process(self, tnode: "TerminalNode") -> None:
        """Post-process the receiver type on behalf of a terminal node.

        By default do nothing.
        """
        pass

    @classmethod
    def _resolve_type(cls, stmt: Statement, sctx: SchemaContext) -> "DataType":
        typ = cast(str, stmt.argument)
        if typ in cls.dtypes:
            res = cls.dtypes[typ](sctx, None)
            res._handle_properties(stmt, sctx)
        else:
            p, s, loc = typ.partition(":")
            res = cls._derived_type(stmt, sctx, loc if s else typ)
        return res

    @classmethod
    def _derived_type(cls, stmt: Statement, sctx: SchemaContext,
                      name: YangIdentifier) -> "DataType":
        tchain = []
        ts = stmt
        sc = sctx
        while True:
            tdef, sc = sctx.schema_data.get_definition(ts, sc)
            ts = cast(Statement, tdef.find1("type", required=True))
            tchain.append((tdef, ts, sc))
            if ts.argument in cls.dtypes:
                break
        res = cls.dtypes[ts.argument](sctx, name)
        btyp = True
        while tchain:
            tdef, typst, tsc = tchain.pop()
            if btyp:
                res._handle_properties(ts, sc)
                btyp = False
            else:
                res._handle_restrictions(typst, tsc)
            ust = tdef.find1("units")
            if ust:
                res.units = ust.argument
            dfst = tdef.find1("default")
            if dfst:
                dval = cast(str, dfst.argument)
                res.default = res.from_yang(dval)
                if res.default is None:
                    raise InvalidArgument(dval)
        res._handle_restrictions(stmt, sctx)
        return res

    def _handle_properties(self, stmt: Statement, sctx: SchemaContext) -> None:
        """Handle type substatements."""
        self._handle_restrictions(stmt, sctx)

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        """Handle type restriction substatements."""
        pass

    def _type_digest(self, config: bool) -> dict[str, Any]:
        """Return receiver's type digest.

        Args:
            config: Specifies whether the type is on a configuration node.
        """
        res = {"base": self.yang_type()}
        if self.name is not None:
            res["derived"] = self.name
        return res


class EmptyType(DataType[tuple[None], list[None]]):
    """Class representing YANG "empty" type."""

    def __contains__(self, val: tuple[None]) -> bool:
        if val == (None,):
            return True
        self._set_error_info()
        return False

    def canonical_string(self, val: tuple[None]) -> Optional[str]:
        return ""

    def parse_value(self, text: str) -> Optional[tuple[None]]:
        return (None,) if text == "" else None

    def from_raw(self, raw: list[None]) -> Optional[tuple[None]]:
        return (None,) if raw == [None] else None

    def to_raw(self, val: tuple[None]) -> list[None]:
        return [None]

    def from_xml(self, xml: ET.Element) -> Optional[tuple[None]]:
        return (None,) if xml.text is None else None

    def to_xml(self, val: tuple[None]) -> None:
        return None


class BitsType(DataType[tuple[str, ...], str]):
    """Class representing YANG "bits" type."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.bit: dict[str, int] = {}

    def __contains__(self, val: tuple[str, ...]) -> bool:
        for b in val:
            if b not in self.bit:
                self._set_error_info(error_message="unknown bit " + b)
                return False
        return True

    def sorted_bits(self) -> list[tuple[str, int]]:
        """Return list of bit items sorted by position."""
        return sorted(self.bit.items(), key=lambda x: x[1])

    def from_raw(self, raw: str) -> Optional[tuple[str, ...]]:
        try:
            return tuple(raw.split())
        except AttributeError:
            return None

    def parse_value(self, text: str) -> Optional[tuple[str, ...]]:
        return self.from_raw(text)

    def from_xml(self, xml: ET.Element) -> Optional[tuple[str, ...]]:
        try:
            return tuple(cast(str, xml.text).split())
        except AttributeError:
            return None

    def to_raw(self, val: tuple[str, ...]) -> Optional[str]:
        return self.canonical_string(val)

    def to_xml(self, val: tuple[str, ...]) -> Optional[str]:
        return self.canonical_string(val)

    def as_int(self, val: tuple[str, ...]) -> Optional[int]:
        """Transform a "bits" value to an integer."""
        res = 0
        try:
            for b in val:
                res += 1 << self.bit[b]
        except KeyError:
            return None
        return res

    def canonical_string(self, val: tuple[str, ...]) -> Optional[str]:
        try:
            items = [(self.bit[b], b) for b in val]
        except KeyError:
            return None
        items.sort()
        return " ".join([x[1] for x in items])

    def _handle_properties(self, stmt: Statement,
                           sctx: SchemaContext) -> None:
        """Handle **bit** statements."""
        nextpos = 0
        for bst in stmt.find_all("bit"):
            if not sctx.schema_data.if_features(bst, sctx.text_mid):
                continue
            label = cast(str, bst.argument)
            pst = bst.find1("position")
            if pst:
                pos = int(cast(str, pst.argument))
                self.bit[label] = pos
                if pos > nextpos:
                    nextpos = pos
            else:
                self.bit[label] = nextpos
            nextpos += 1

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        bst = stmt.find_all("bit")
        if not bst:
            return
        new = set([b.argument for b in bst if
                   sctx.schema_data.if_features(b, sctx.text_mid)])
        for bit in set(self.bit) - new:
            del self.bit[bit]

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        bits: list[Optional[str]] = []
        i = 0
        for b in self.sorted_bits():
            while i < b[1]:
                bits.append(None)
                i += 1
            bits.append(b[0])
        res["bits"] = bits
        return res


class BooleanType(DataType[bool, bool]):
    """Class representing YANG "boolean" type."""

    def __contains__(self, val: bool) -> bool:
        if isinstance(val, bool):
            return True
        self._set_error_info()
        return False

    def from_raw(self, raw: bool) -> Optional[bool]:
        return raw if isinstance(raw, bool) else None

    def from_xml(self, xml: ET.Element) -> Optional[bool]:
        """Return a cooked value of the received XML type.

        Args:
            xml: Text of the XML node
        """
        return self.parse_value(cast(str, xml.text))

    def to_raw(self, val: bool) -> Optional[bool]:
        return val if val in self else None

    def parse_value(self, text: str) -> Optional[bool]:
        """Parse boolean value.

        Args:
            text: String representation of the value.
        """
        if text == "true":
            return True
        elif text == "false":
            return False
        else:
            return None

    def canonical_string(self, val: bool) -> Optional[str]:
        if val is True:
            return "true"
        elif val is False:
            return "false"
        else:
            return None

    def to_xml(self, val: bool) -> Optional[str]:
        return self.canonical_string(val)


class LinearType(DataType[L, str]):
    """Abstract class representing character or byte sequences."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.length: Optional[Intervals[int]] = None

    def __contains__(self, val: L) -> bool:
        if self.length and len(val) not in self.length:
            self._set_error_info(self.length.error_tag,
                                 self.length.error_message)
            return False
        return True

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        lstmt = stmt.find1("length")
        if lstmt:
            if self.length is None:
                self.length = Intervals[int](
                    [[0, 4294967295]], error_message="invalid length")
            self.length.restrict_with(
                cast(str, lstmt.argument), *lstmt.get_error_info())

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if self.length:
            res["length"] = self.length.intervals
        return res


class StringType(LinearType[str]):
    """Class representing YANG "string" type."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.patterns: list[Pattern] = []

    def __contains__(self, val: str) -> bool:
        if not isinstance(val, str):
            self._set_error_info()
            return False
        if not super().__contains__(val):
            return False
        for p in self.patterns:
            if (p.regex.match(val) is not None) == p.invert_match:
                self._set_error_info(
                    p.error_tag, cast(str, p.error_message) + ': ' + val)
                return False
        return True

    def from_raw(self, raw: str) -> Optional[str]:
        return raw if isinstance(raw, str) else None

    def from_xml(self, xml: ET.Element) -> Optional[str]:
        return self.from_raw(cast(str, xml.text))

    def to_raw(self, val: str) -> Optional[str]:
        return val if val in self else None

    def parse_value(self, text: str) -> Optional[str]:
        return self.from_raw(text)

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        super()._handle_restrictions(stmt, sctx)
        for pst in stmt.find_all("pattern"):
            invm = pst.find1("modifier", "invert-match") is not None
            self.patterns.append(Pattern(
                cast(str, pst.argument), invm, *pst.get_error_info()))

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        pats = [p.pattern for p in self.patterns if not p.invert_match]
        ipats = [p.pattern for p in self.patterns if p.invert_match]
        if pats:
            res["patterns"] = pats
        if ipats:
            res["neg_patterns"] = ipats
        return res


class BinaryType(LinearType[bytes]):
    """Class representing YANG "binary" type."""

    def __contains__(self, val: bytes) -> bool:
        if not isinstance(val, bytes):
            self._set_error_info()
            return False
        return super().__contains__(val)

    def from_raw(self, raw: str) -> Optional[bytes]:
        """Override superclass method."""
        try:
            return base64.b64decode(raw, validate=True)
        except (TypeError, binascii.Error):
            return None

    def parse_value(self, text: str) -> Optional[bytes]:
        return self.from_raw(text)

    def from_xml(self, xml: ET.Element) -> Optional[bytes]:
        """Override superclass method."""
        try:
            return base64.b64decode(cast(str, xml.text), validate=True)
        except TypeError:
            return None

    def to_raw(self, val: bytes) -> Optional[str]:
        return self.canonical_string(val)

    def to_xml(self, val: bytes) -> Optional[str]:
        return self.canonical_string(val)

    def canonical_string(self, val: bytes) -> Optional[str]:
        try:
            return base64.b64encode(val).decode("ascii")
        except TypeError:
            return None


class EnumerationType(DataType[str, str]):
    """Class representing YANG "enumeration" type."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.enum: dict[str, int] = {}

    def __contains__(self, val: str) -> bool:
        if val in self.enum:
            return True
        self._set_error_info()
        return False

    def from_raw(self, raw: str) -> Optional[str]:
        return raw if isinstance(raw, str) else None

    def from_xml(self, xml: ET.Element) -> Optional[str]:
        return self.from_raw(cast(str, xml.text))

    def to_raw(self, val: str) -> Optional[str]:
        return val if val in self else None

    def parse_value(self, text: str) -> Optional[str]:
        return self.from_raw(text)

    def sorted_enums(self) -> list[tuple[str, int]]:
        """Return list of enum items sorted by value."""
        return sorted(self.enum.items(), key=lambda x: x[1])

    def _handle_properties(self, stmt: Statement,
                           sctx: SchemaContext) -> None:
        """Handle **enum** statements."""
        nextval = 0
        for est in stmt.find_all("enum"):
            if not sctx.schema_data.if_features(est, sctx.text_mid):
                continue
            label = cast(str, est.argument)
            vst = est.find1("value")
            if vst:
                val = int(cast(str, vst.argument))
                self.enum[label] = val
                if val > nextval:
                    nextval = val
            else:
                self.enum[label] = nextval
            nextval += 1

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        est = stmt.find_all("enum")
        if not est:
            return
        new = set([e.argument for e in est if
                   sctx.schema_data.if_features(e, sctx.text_mid)])
        for en in set(self.enum) - new:
            del self.enum[en]

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if config:
            res["enums"] = list(self.enum.keys())
        return res


class LinkType(DataType[S, RS]):
    """Abstract class for instance-referencing types."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.require_instance: bool = True

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        if stmt.find1("require-instance", "false"):
            self.require_instance = False


class LeafrefType(LinkType[ScalarValue, RawScalar]):
    """Class representing YANG "leafref" type."""

    path: Expr
    ref_type: DataType

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)

    def __contains__(self, val: ScalarValue) -> bool:
        return val in self.ref_type

    def _handle_properties(self, stmt: Statement, sctx: SchemaContext) -> None:
        super()._handle_properties(stmt, sctx)
        pathst = cast(Statement, stmt.find1("path", required=True))
        self.path = XPathParser(cast(str, pathst.argument), sctx).parse()

    def canonical_string(self, val: ScalarValue) -> Optional[str]:
        return self.ref_type.canonical_string(val)

    def from_raw(self, raw: RawScalar) -> Optional[ScalarValue]:
        return self.ref_type.from_raw(raw)

    def from_xml(self, xml: ET.Element) -> Optional[bytes]:
        return self.ref_type.from_xml(xml)

    def to_raw(self, val: ScalarValue) -> Optional[RawScalar]:
        return self.ref_type.to_raw(val)

    def to_xml(self, val: ScalarValue) -> Optional[str]:
        return self.ref_type.to_xml(val)

    def parse_value(self, text: str) -> Optional[ScalarValue]:
        return self.ref_type.parse_value(text)

    def from_yang(self, text: str) -> ScalarValue:
        return self.ref_type.from_yang(text)

    def _deref(self, node: InstanceNode) -> list[InstanceNode]:
        ns = self.path.evaluate(node)
        if isinstance(ns, NodeSet):
            return [n for n in ns if str(n) == str(node)]
        raise XPathTypeError(str(ns))

    def _post_process(self, tnode: "TerminalNode") -> None:
        ref = tnode._follow_leafref(self.path, tnode)
        if ref is None:
            raise InvalidLeafrefPath(tnode.qual_name)
        self.ref_type = ref.type

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        res["ref_type"] = self.ref_type._type_digest(config)
        return res


class InstanceIdentifierType(LinkType[InstanceRoute, InstanceIdentifier]):
    """Class representing YANG "instance-identifier" type."""

    def __str__(self):
        return "instance-identifier"

    def __contains__(self, val: InstanceRoute) -> bool:
        return isinstance(val, InstanceRoute)

    def yang_type(self) -> YangIdentifier:
        """Override the superclass method."""
        return "instance-identifier"

    def from_raw(self, raw: InstanceIdentifier) -> Optional[InstanceRoute]:
        try:
            return InstanceIdParser(raw).parse()
        except ParserException:
            return None

    def parse_value(self, text: str) -> Optional[InstanceRoute]:
        return self.from_raw(text)

    def from_xml(self, xml: ET.Element) -> Optional[InstanceRoute]:
        return self.from_raw(cast(InstanceIdentifier, xml.text))

    def to_raw(self, val: InstanceRoute) -> InstanceIdentifier:
        """Override the superclass method."""
        return str(val)

    def to_xml(self, val: InstanceRoute) -> InstanceIdentifier:
        """Override the superclass method."""
        return str(val)

    def from_yang(self, text: str) -> InstanceRoute:
        """Override the superclass method."""
        return XPathParser(text, self.sctx).parse().as_instance_route()

    @staticmethod
    def _deref(node: InstanceNode) -> list[InstanceNode]:
        return [node.top().goto(cast(InstanceRoute, node.value))]


class IdentityrefType(DataType[QualName, YangIdentifier]):
    """Class representing YANG "identityref" type."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.bases: list[QualName] = []

    def __contains__(self, val: QualName) -> bool:
        for b in self.bases:
            if not self.sctx.schema_data.is_derived_from(val, b):
                self._set_error_info(
                    error_message=f"'{self.canonical_string(val)}'" +
                    f" not derived from '{self.canonical_string(b)}'")
                return False
        return True

    def from_raw(self, raw: YangIdentifier) -> Optional[QualName]:
        try:
            i1, s, i2 = raw.partition(":")
        except AttributeError:
            return None
        return (i2, i1) if s else (i1, self.sctx.default_ns)

    def parse_value(self, text: str) -> Optional[QualName]:
        return self.from_raw(text)

    def from_xml(self, xml: ET.Element) -> Optional[QualName]:
        try:
            i1, s, i2 = cast(str, xml.text).partition(":")
        except AttributeError:
            return None

        if not s:
            # not prefixed, so assume default ns
            return (i1, self.sctx.default_ns)
        elif i1 == self.sctx.default_ns:
            # prefixed with default ns
            return (i2, self.sctx.default_ns)

        # the following code has issues (Issue #79)
        ns_url = xml.attrib.get('xmlns:'+i1)
        if not ns_url:
            raise MissingModuleNamespace(cast(str, ns_url))
        module = self.sctx.schema_data.modules_by_ns.get(ns_url)
        if not module:
            raise MissingModuleNamespace(ns_url)
        return (i2, module.main_module[0])

    def to_raw(self, val: QualName) -> YangIdentifier:
        return self.canonical_string(val)

    def to_xml(self, val: QualName) -> YangIdentifier:
        return self.canonical_string(val)

    def from_yang(self, text: YangIdentifier) -> QualName:
        """Override the superclass method."""
        try:
            return self.sctx.schema_data.translate_pname(
                text, self.sctx.text_mid)
        except (ModuleNotRegistered, UnknownPrefix):
            raise InvalidArgument(text)

    def canonical_string(self, val: QualName) -> str:
        """Return canonical form of a value."""
        return f"{val[1]}:{val[0]}"

    def _handle_properties(self, stmt: Statement,
                           sctx: SchemaContext) -> None:
        self.bases = []
        for b in stmt.find_all("base"):
            self.bases.append(
                sctx.schema_data.translate_pname(
                    cast(str, b.argument), sctx.text_mid))

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if config:
            res["identities"] = list(
                self.sctx.schema_data.derived_from_all(self.bases))
        return res


class NumericType(DataType[N, RN]):
    """Abstract class for numeric data types."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.range: Optional[Intervals[N]] = None

    def __contains__(self, val: N) -> bool:
        if self.range is None:
            if self._range[0] <= val <= self._range[1]:
                return True
            else:
                self._set_error_info()
                return False
        if val in self.range:
            return True
        self._set_error_info(self.range.error_tag, self.range.error_message)
        return False

    @property
    @abstractmethod
    def _range(self):
        """List with minimum and maximum value permitted by the type."""

    def _handle_restrictions(self, stmt: Statement,
                             sctx: SchemaContext) -> None:
        rstmt = stmt.find1("range")
        if rstmt:
            if self.range is None:
                self.range = Intervals[N](
                    [self._range], parser=self.parse_value,
                    error_message="not in range")
            self.range.restrict_with(
                cast(str, rstmt.argument), *rstmt.get_error_info())

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        if self.range:
            res["range"] = [[self.to_raw(r[0]), self.to_raw(r[-1])]
                            for r in self.range.intervals]
        return res


class Decimal64Type(NumericType[decimal.Decimal, str]):
    """Class representing YANG "decimal64" type."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self._epsilon = decimal.Decimal(0)

    def __contains__(self, val: decimal.Decimal) -> bool:
        if not isinstance(val, decimal.Decimal):
            self._set_error_info()
            return False
        return super().__contains__(val)

    @property
    def _range(self) -> list[decimal.Decimal]:
        quot = decimal.Decimal(10**self.fraction_digits)
        lim = decimal.Decimal(9223372036854775808)
        return [-lim / quot, (lim - 1) / quot]

    def _handle_properties(self, stmt: Statement, sctx: SchemaContext) -> None:
        fd = cast(Statement, stmt.find1("fraction-digits", required=True))
        self.fraction_digits = int(cast(str, fd.argument))
        self._epsilon = decimal.Decimal(10) ** -self.fraction_digits
        super()._handle_properties(stmt, sctx)

    def from_raw(self, raw: str) -> Optional[decimal.Decimal]:
        """Override superclass method.

        According to [RFC7951]_, a raw instance must be string.
        """
        try:
            return (decimal.Decimal(raw).quantize(self._epsilon)
                    if isinstance(raw, str) else None)
        except decimal.InvalidOperation:
            return None

    def parse_value(self, text: str) -> Optional[decimal.Decimal]:
        return self.from_raw(text)

    def from_xml(self, xml: ET.Element) -> Optional[decimal.Decimal]:
        try:
            return (decimal.Decimal(xml.text).quantize(self._epsilon)
                    if isinstance(xml.text, str) else None)
        except decimal.InvalidOperation:
            return None

    def to_raw(self, val: decimal.Decimal) -> str:
        return self.canonical_string(val)

    def to_xml(self, val: decimal.Decimal) -> str:
        return self.canonical_string(val)

    def canonical_string(self, val: decimal.Decimal) -> str:
        if val == 0:
            return "0.0"
        sval = str(val.quantize(self._epsilon)).rstrip("0")
        return (sval + "0") if sval.endswith(".") else sval

    def _type_digest(self, config: bool) -> dict[str, Any]:
        res = super()._type_digest(config)
        res["fraction_digits"] = self.fraction_digits
        return res


class IntegralType(NumericType[int, RN]):
    """Abstract class for integral data types."""

    octhex = re.compile("[-+]?0([x0-9])")
    """Regular expression for octal or hexadecimal default."""

    def __contains__(self, val: int) -> bool:
        if not isinstance(val, int) or isinstance(val, bool):
            self._set_error_info()
            return False
        return super().__contains__(val)

    def parse_value(self, text: str) -> Optional[int]:
        """Override superclass method."""
        try:
            return int(text)
        except ValueError:
            return None

    def from_raw(self, raw: RN) -> Optional[int]:
        if not isinstance(raw, int) or isinstance(raw, bool):
            return None
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def from_xml(self, xml: ET.Element) -> Optional[int]:
        try:
            return int(xml.text) if isinstance(xml.text, str) else None
        except (ValueError, TypeError):
            return None

    def from_yang(self, text: str) -> int:
        """Override the superclass method."""
        mo = self.octhex.match(text)
        if mo:
            base = 16 if mo.group(1) == "x" else 8
        else:
            base = 10
        try:
            return int(text, base)
        except (ValueError, TypeError):
            raise InvalidArgument(text)


class Int8Type(IntegralType[int]):
    """Class representing YANG "int8" type."""

    _range = [-128, 127]

    def to_raw(self, val: int) -> Optional[int]:
        return val if val in self else None


class Int16Type(IntegralType[int]):
    """Class representing YANG "int16" type."""

    _range = [-32768, 32767]

    def to_raw(self, val: int) -> Optional[int]:
        return val if val in self else None


class Int32Type(IntegralType[int]):
    """Class representing YANG "int32" type."""

    _range = [-2147483648, 2147483647]

    def to_raw(self, val: int) -> Optional[int]:
        return val if val in self else None


class Int64Type(IntegralType[str]):
    """Class representing YANG "int64" type."""

    _range = [-9223372036854775808, 9223372036854775807]

    def from_raw(self, raw: str) -> Optional[int]:
        """Override superclass method.

        According to [RFC7951]_, a raw instance must be string.
        """
        if not isinstance(raw, str) or isinstance(raw, bool):
            return None
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def from_xml(self, xml: ET.Element) -> Optional[int]:
        """Override superclass method.

        XML is always delivered as a text element
        """
        try:
            return int(xml.text) if isinstance(xml.text, str) else None
        except (ValueError, TypeError):
            return None

    def to_raw(self, val: int) -> Optional[str]:
        return self.canonical_string(val)

    def to_xml(self, val: int) -> Optional[str]:
        return self.canonical_string(val)


class Uint8Type(IntegralType[int]):
    """Class representing YANG "uint8" type."""

    _range = [0, 255]

    def to_raw(self, val: int) -> Optional[int]:
        return val if val in self else None


class Uint16Type(IntegralType[int]):
    """Class representing YANG "uint16" type."""

    _range = [0, 65535]

    def to_raw(self, val: int) -> Optional[int]:
        return val if val in self else None


class Uint32Type(IntegralType[int]):
    """Class representing YANG "uint32" type."""

    _range = [0, 4294967295]

    def to_raw(self, val: int) -> Optional[int]:
        return val if val in self else None


class Uint64Type(IntegralType[str]):
    """Class representing YANG "uint64" type."""

    _range = [0, 18446744073709551615]

    def from_raw(self, raw: str) -> Optional[int]:
        """Override superclass method.

        According to [RFC7951]_, a raw instance must be string.
        """
        if not isinstance(raw, str):
            return None
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def from_xml(self, xml: ET.Element) -> Optional[int]:
        """Override superclass method.

        XML is always delivered as a text element
        """
        try:
            return int(xml.text) if isinstance(xml.text, str) else None
        except (ValueError, TypeError):
            return None

    def to_raw(self, val: int) -> Optional[str]:
        return self.canonical_string(val)

    def to_xml(self, val: int) -> Optional[str]:
        return self.canonical_string(val)


class UnionType(DataType[ScalarValue, RawScalar]):
    """Class representing YANG "union" type."""

    def __init__(self, sctx: SchemaContext, name: YangIdentifier):
        """Initialize the class instance."""
        super().__init__(sctx, name)
        self.types: list[DataType] = []

    def __contains__(self, val: ScalarValue) -> bool:
        for t in self.types:
            try:
                if val in t:
                    return True
            except TypeError:
                continue
        return False

    def to_raw(self, val: ScalarValue) -> Optional[RawScalar]:
        for t in self.types:
            if val in t:
                return t.to_raw(val)
        return None

    def to_xml(self, val: ScalarValue) -> Optional[str]:
        for t in self.types:
            if val in t:
                return t.to_xml(val)
        return None

    def canonical_string(self, val: ScalarValue) -> Optional[str]:
        for t in self.types:
            if val in t:
                return t.canonical_string(val)
        return None

    def parse_value(self, text: str) -> Optional[ScalarValue]:
        for t in self.types:
            val = t.parse_value(text)
            if val is not None and val in t:
                return val
        return None

    def from_raw(self, raw: RawScalar) -> Optional[ScalarValue]:
        for t in self.types:
            val = t.from_raw(raw)
            if val is not None and val in t:
                return val
        return None

    def from_xml(self, xml: ET.Element) -> Optional[ScalarValue]:
        for t in self.types:
            val = t.from_xml(xml)
            if val is not None and val in t:
                return val
        return None

    def _handle_properties(self, stmt: Statement,
                           sctx: SchemaContext) -> None:
        self.types = [self._resolve_type(ts, sctx)
                      for ts in stmt.find_all("type")]

    def _post_process(self, tnode: "TerminalNode") -> None:
        for t in self.types:
            t._post_process(tnode)


DataType.dtypes = {"binary": BinaryType,
                   "bits": BitsType,
                   "boolean": BooleanType,
                   "decimal64": Decimal64Type,
                   "empty": EmptyType,
                   "enumeration": EnumerationType,
                   "identityref": IdentityrefType,
                   "instance-identifier": InstanceIdentifierType,
                   "int8": Int8Type,
                   "int16": Int16Type,
                   "int32": Int32Type,
                   "int64": Int64Type,
                   "leafref": LeafrefType,
                   "string": StringType,
                   "uint8": Uint8Type,
                   "uint16": Uint16Type,
                   "uint32": Uint32Type,
                   "uint64": Uint64Type,
                   "union": UnionType
                   }
"""Dictionary mapping type names to classes."""



# Copyright © 2016–2026 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""A route into an instance value that is also used as the cooked value
of the instance-identifier type.

This module implements the following classes:

* InstanceRouteItem: Protocol class defining API for instance route items.
* InstanceRoute: Route into an instance value.
"""
from typing import Optional, Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from .instance import InstanceNode
    from .instvalue import StructuredValue, Value
    from .schemanode import DataNode


class InstanceRouteItem(Protocol):
    """This protocol class defines a required API for instance route items."""

    def __eq__(self, other: object) -> bool:
        """Return True if the receiver is equal to other."""

    def __str__(self) -> str:
        """Return string representation of the receiver (i-i segment)."""

    def peek_step(self, val: "StructuredValue",
                  sn: "DataNode") -> tuple[Optional["Value"], "DataNode"]:
        """Return value addressed by the receiver relative to the current
        position together with its schema node.

        Args:
            val: Value at the current position.
            sn:  Schema node corresponding to the current position.
        """

    def goto_step(self, inst: "InstanceNode") -> "InstanceNode":
        """Return instance node addressed by the receiver.

        Args:
            inst: Current instance node.
        """


class InstanceRoute(tuple[InstanceRouteItem, ...]):
    """This class represents a route into an instance value."""

    def __str__(self) -> str:
        """Return instance-id as the string representation of the receiver."""
        if self:
            return "".join([str(c) for c in self])
        else:
            return "/"

    def __hash__(self) -> int:
        """Return the hash value of the receiver."""
        return self.__str__().__hash__()
